// SOURCE CODE HEADER
//
// myservice.cpp
// Purpose: Run an application as a NT/95/98 service
// Contact: jeffg@cips.ca
// Legalese: This source code is public domain
// Based on MSDN example code by Nigel Thomson (nigelt@microsoft.com)
// Modified to also run as Win95/98 service by Steve Porter (steve@dcypher.net)
// Modified for the ECDL2K-108 challenge by Jeff Gilchrist (jeffg@cips.ca)

#include <stdio.h>
#include <windows.h>
#include "myservice.h"

extern int gotSignal;
extern int ecdl_start(int argc, char *argv[], FILE* flog, char *exe_path, int pr);
extern void ecdl_stop(void);  

CMyService::CMyService(void) : CNTService("ECDL_Service"), 
							m_iStartParam(0), m_iIncParam(1), m_iState(0) { }

// Read the registry parameters, try opening the registry key:
// HKEY_LOCAL_MACHINE\SYSTEM\CurrentControlSet\Services\<AppName>\Parameters

BOOL CMyService::OnInit(void)
{	HKEY	hkey;
	char	szKey[1024];

	strcpy(szKey, "SYSTEM\\CurrentControlSet\\Services\\");
	strcat(szKey, m_szServiceName);
	strcat(szKey, "\\Parameters");
    
	if(RegOpenKeyEx(HKEY_LOCAL_MACHINE, szKey, 0, KEY_QUERY_VALUE, &hkey) == ERROR_SUCCESS) 
	{
        DWORD dwType = 0;	// Yes we are installed
        DWORD dwSize = sizeof(m_iStartParam);
        RegQueryValueEx(hkey, "Start", NULL, &dwType, (BYTE*)&m_iStartParam, &dwSize);
        dwSize = sizeof(m_iIncParam);
        RegQueryValueEx(hkey, "Inc", NULL, &dwType, (BYTE*)&m_iIncParam, &dwSize);
        RegCloseKey(hkey);
    }
	
	m_iState = m_iStartParam;	// Set the initial state

    printf("Starting ECDL2K-108 Client\n");

    return TRUE;
}

int get_priority(char *sp)
{	int	pr;

	if(!*sp)

		return -1;

	if(*sp == 'h' || *sp == 'H')

		pr = 22;		// set 10's digit = 2 for HIGH priority

	else if(*sp == 'n' || *sp == 'N')

		pr = 12;		// set 10's digit = 1 for NORMAL priority

	else if(*sp == 'l' || *sp == 'L')

		pr = 2;			// set 10's digit = 0 for LOW (IDLE) priority

	else 
	{
		pr = 2; sp--;	// 1's digit = 2 is normal priority within class
	}

	sp++;

	if(!*sp || *sp == '0' && !*(sp + 1))

		return pr;		// if no priority modifier or the modifier is '0'

	if(*sp == '+')			// increase the priority
	{
		if(!*++sp)						// one '+' character  -> +1

			return pr + 1;

		if(*sp == '+' && !*(sp + 1))	// two '+' characters -> +2

			return pr + 2;

		else if(*sp >= '0' && *sp <= '2' && !*(sp + 1))

			return pr + *sp - '0';		// or '+' with '0', '1' or '2'
	}
	else if(*sp == '-')		// reduce the priority
	{
		if(!*++sp)						// one '-' character  -> -1

			return pr - 1;

		if(*sp == '-' && !*(sp + 1))	// two '-' characters -> -2

			return pr - 2;

		else if(*sp >= '0' && *sp <= '2' && !*(sp + 1))

			return pr - *sp + '0';		// or '-' with '0', '1' or '2'
	}

	return -1;
}

// code runs here when service starts

void CMyService::Run(void)
{	FILE	*flog, *fini;

	char	mode[_MAX_PATH] = {"http"};
	char	machine[_MAX_PATH] = {"MACHINE"};
	char	email[_MAX_PATH] = {"ADDRESS"};
	char	proxy[_MAX_PATH] = {"0"};
	int		pr = 0;					// default thread priority 

    char	szFilePath[_MAX_PATH];	// file pathnames
    char	szIniPath[_MAX_PATH];
    char	szLogPath[_MAX_PATH];
	
    // Set the file paths using the executable file path as a basis

    GetModuleFileName(NULL, szFilePath, sizeof(szFilePath));
	strcpy(szIniPath, szFilePath); strcpy(szIniPath + strlen(szIniPath) - 3, "ini");
	strcpy(szLogPath, szFilePath); strcpy(szLogPath + strlen(szLogPath) - 3, "log");

	// open the LOG file for writing
	
	if((flog = fopen(szLogPath, "a+")) == NULL)
	{
		DebugMsg("Could not open LOG file for writing"); exit(1);
	}

	// open the INI file for reading

	if((fini = fopen(szIniPath, "r")) == NULL)
	{
		DebugMsg("Could not open INI file for reading. Halting!");
		fputs("Could not open INI file for reading. Halting!\n", flog);
		fclose(flog); exit(1);
	}

	char	*line  = new char[_MAX_PATH + 1];
	char	*dp, *sp;
	int		line_no = 0;
	
	while(fgets(line, _MAX_PATH + 1, fini) != NULL)
	{
		line_no++;	

		if(line[0] == '#' || line[0] == '\n')	// check if line is a comment

			continue;	// if line is empty or a comment

		if(line[(strlen(line) - 1)] == '\n')	// get rid of newline character

			line[(strlen(line) - 1)] = '\0';

		if(strncmp(line, "MODE=", 5) == 0)
		{
			dp = mode; sp = line + 5;
		}
		else if(strncmp(line,"MACHNAME=",9) == 0)
		{
			dp = machine; sp = line + 9;
		}
		else if(strncmp(line,"EMAIL=",6) == 0)
		{
			dp = email; sp = line + 6;
		}
		else if(strncmp(line,"PROXY=",6) == 0)
		{
			dp = proxy; sp = line + 6;
		}
		else if(strncmp(line,"PRIORITY=",9) == 0)
		{
			if((pr = get_priority(line + 9)) == -1)
			{
				fputs("** WARNING: INI file line \'", flog); fputs(line, flog);
				sprintf(line, "\' (line %d) is not understood **\n", line_no); 
				fputs(line, flog); pr = 0;
			}

			continue;
		}
		else
		{
			fputs("** WARNING: INI file line \'", flog); fputs(line, flog);
			sprintf(line, "\' (line %d) is not understood **\n", line_no); 
			fputs(line, flog); continue;
		}

		if(*sp)

			strcpy(dp, sp);
	}

	fclose(fini); fflush(flog); delete[]line;

	// package up parameters in argc/argv[] format to send to ECDL client

	int		parmc;
	char	*parmv[8];

	parmc = (proxy[0] == '0' ? 6 : 8);
	
	parmv[0] = new char[strlen(szFilePath) + 1]; strcpy(parmv[0], szFilePath);
	parmv[1] = new char[strlen(mode) + 1]; strcpy(parmv[1], mode);
	parmv[2] = new char[3]; strcpy(parmv[2], "by");
	parmv[3] = new char[strlen(email) + 1]; strcpy(parmv[3], email);
	parmv[4] = new char[3]; strcpy(parmv[4], "on");
	parmv[5] = new char[strlen(machine) + 1]; strcpy(parmv[5], machine);

	if(parmc == 8)	// if proxy is set there must be a parameter
	{
		parmv[6] = new char[6]; strcpy(parmv[6], "proxy");
		parmv[7] = new char[strlen(proxy) + 1]; strcpy(parmv[7], proxy);
	}

	if(ecdl_start(parmc, parmv, flog, szFilePath, pr))
	{
	    while(m_bIsRunning && !gotSignal)	// loop while service is running
		{
			m_iState += m_iIncParam;		//update the current state

			Sleep(5000L);
	    }

		ecdl_stop();
	}
	else

		fputs("Error during service startup\n", flog);

	fclose(flog);

	delete[]parmv[0]; delete[]parmv[1];
	delete[]parmv[2]; delete[]parmv[3];
	delete[]parmv[4]; delete[]parmv[5];

	if(parmc == 8)
	{
		delete[]parmv[6]; delete[]parmv[7];
	}
}

// Process user control requests

BOOL CMyService::OnUserControl(DWORD dwOpcode)
{
    switch(dwOpcode) 
	{
		case SERVICE_CONTROL_USER + 0:	// Save the current status in the registry
			SaveStatus(); return TRUE;
		default: 						// say not handled
			return FALSE;
    }
}

void CMyService::OnShutdown(void)
{
	ecdl_stop();
}

// Save the current status in the registry

void CMyService::SaveStatus(void)
{	HKEY	hkey = NULL;
	char	szKey[1024];
    DWORD	dwDisp;
	DWORD	dwErr;

    DebugMsg("Saving current status");

    // Try opening the registry key:
    // HKEY_LOCAL_MACHINE\SYSTEM\CurrentControlSet\Services\<AppName>\...

	strcpy(szKey, "SYSTEM\\CurrentControlSet\\Services\\");
	strcat(szKey, m_szServiceName);
	strcat(szKey, "\\Status");

    DebugMsg("Creating key: %s", szKey);

    dwErr = RegCreateKeyEx(	HKEY_LOCAL_MACHINE, szKey, 0, "", REG_OPTION_NON_VOLATILE,
                   			KEY_WRITE, NULL, &hkey, &dwDisp);
	
	if(dwErr != ERROR_SUCCESS) 
	{
		DebugMsg("Failed to create Status key (%lu)", dwErr); return;
	}	

	DebugMsg("Saving 'Current' as %ld", m_iState);	// Set the registry values
    RegSetValueEx(hkey, "Current", 0, REG_DWORD, (BYTE*)&m_iState, sizeof(m_iState));

    RegCloseKey(hkey);
}
